local pairs, ipairs, find, type, collectgarbage, insert, getn, remove = pairs, ipairs, string.find, type, collectgarbage, table.insert, getn, table.remove;
local wipe = table.wipe
local ScrollMulti = 4

local function ReleaseElements(self)
    for i,v in ipairs(self.ElementNames) do 
        if self["Container_Element_" .. v] then 
            self["Container_Element_" .. v]:Release()
            self["Container_Element_" .. v] = nil
        end
    end 

    wipe(self.ElementNames)
    self.NumDisplayItems = 0
    wipe(self.ContainerRowData)
end

local function ClearItems(self)
    local i = 1
    while self["Container_Element_" .. i] do 
        self["Container_Element_" .. i]:Release()
        self["Container_Element_" .. i] = nil
        i = i + 1
    end 

    self.NumDisplayItems = 0
    wipe(self.ContainerRowData)
    self.Content:Size(nil,1)
end 

local tempDataStore = {}
local function AddRow(self,...)
    wipe(tempDataStore)
    local i = 1
    local CurrentItemIndex = 0
    local VarName,VarValue = select((2*i - 1),...)
    local HeightOffset = 0
    while type(VarName) == "string" do 
        if VarName == "HeightOffset" then 
            HeightOffset = VarValue
        elseif VarName == "Header" then 
            self:AddRow("Type","Line","HeightOffset",20)
            self:AddRow("Type","LB","Text",VarValue or "")
            return 
        end 
        if select((2*i - 1),...) == "Type" then
            insert(tempDataStore,{})
            CurrentItemIndex = CurrentItemIndex + 1
        end
        if tempDataStore[CurrentItemIndex] then 
            tempDataStore[CurrentItemIndex][select((2*i - 1),...)] = select(2*i,...)
        end
        i = i + 1
        VarName,VarValue = select((2*i - 1),...)
    end

    local TotalElements = getn(tempDataStore)
    local Ratio = 1
    local RatioOffset = 0
    local UndefinedWidth = TotalElements
    for i,v in ipairs(tempDataStore) do 
        if v.FixedWidth then 
            RatioOffset = RatioOffset + v.FixedWidth
            UndefinedWidth = UndefinedWidth - 1
        elseif v.WidthRatio then 
            Ratio = Ratio - v.WidthRatio
            UndefinedWidth = UndefinedWidth - 1
        end 
    end 
    if UndefinedWidth > 0 then 
        Ratio = Ratio/UndefinedWidth
    end 

    local PrevFrame 
    if getn(tempDataStore) > 0 then
        insert(self.ContainerRowData,{Height = 0, HeightOffset = HeightOffset, ElementWidthModifier = {}})
        local CurrentRowIndex = getn(self.ContainerRowData)
        for i,v in ipairs(tempDataStore) do 
            v.Name = v.Name or (self.NumDisplayItems + 1) 

            local Width = self.Content:GetWidth() - self.self.Padding * (TotalElements + 1)
            self.ContainerRowData[CurrentRowIndex].ElementWidthModifier[i] = self.ContainerRowData[CurrentRowIndex].ElementWidthModifier[i] or {Name = v.Name}
            if v.FixedWidth then 
                Width = v.FixedWidth 
                self.ContainerRowData[CurrentRowIndex].ElementWidthModifier[i].FixedWidth = v.FixedWidth
            else 
                Width = (v.WidthRatio or Ratio)*(Width - RatioOffset)
                self.ContainerRowData[CurrentRowIndex].ElementWidthModifier[i].RatioOffset = RatioOffset
                self.ContainerRowData[CurrentRowIndex].ElementWidthModifier[i].Ratio = (v.WidthRatio or Ratio)
            end 

            if v.Type == "DB" then
                self["Container_Element_" .. v.Name] = self.self:CreateButton(self.Content,v.Borderless)
                self["Container_Element_" .. v.Name]:Size(Width, 20 + (((v.Lines or 1) - 1) * 12))
                if type(v.Text) == "table" then 
                    self["Container_Element_" .. v.Name]:SetText(v.Text)
                else 
                    self["Container_Element_" .. v.Name]:SetText(v.Text or "", 11, "Normal", nil, nil, v.Alignment)
                end 
                if v.Icon then
                    if v.IconData then
                        self["Container_Element_" .. v.Name]:SetIcon("TexturePath",v.Icon,"LockAspectRatio","LEFT",unpack(v.IconData))
                    else
                        self["Container_Element_" .. v.Name]:SetIcon("TexturePath",v.Icon,"LockAspectRatio","LEFT")
                    end
                    self["Container_Element_" .. v.Name].Text1:SetPoint("BOTTOMLEFT",self["Container_Element_" .. v.Name].Icon,"BOTTOMRIGHT", self.self.Padding, self.self.Padding)
                end
                self["Container_Element_" .. v.Name]:SetFocus(v.Selected)
                self["Container_Element_" .. v.Name]:SetDisable(v.Disable)
                self["Container_Element_" .. v.Name].OnClick = v.Func
                if v.CloseBtn then 
                    self["Container_Element_" .. v.Name]:CloseButtonTexture()
                end 
            elseif v.Type == "IT" then 
                self["Container_Element_" .. v.Name] = self.self:FetchTexture(self.Content)
                self["Container_Element_" .. v.Name]:SetSize(Width, 10)
                self["Container_Element_" .. v.Name]:SetTexture(v.Icon)
                self.self:SetThemedColor(self["Container_Element_" .. v.Name],v.Color or "Highlight")
            elseif v.Type == "LB" then
                self["Container_Element_" .. v.Name] = self.self:FetchFontString(self.Content)
                self["Container_Element_" .. v.Name]:SetText(v.Text)
                self["Container_Element_" .. v.Name]:SetWidth(Width)
                if v.FixedHeight then 
                    self["Container_Element_" .. v.Name]:SetHeight(v.FixedHeight)
                end
                self.self:SetThemedColor(self["Container_Element_" .. v.Name],v.Color or "Highlight")
                local Font, _, Flags = self["Container_Element_" .. v.Name]:GetFont()
                self["Container_Element_" .. v.Name]:SetFont(Font, 11, Flags)
                self["Container_Element_" .. v.Name]:SetJustifyH(v.Alignment or "CENTER")
                self["Container_Element_" .. v.Name]:SetJustifyV(v.VAlignment or "CENTER")
            elseif v.Type == "CB" then
                self["Container_Element_" .. v.Name] = self.self:CreateCheckbox(self.Content)
                self["Container_Element_" .. v.Name]:SetText(v.Text, 11, "Normal", nil, nil, "LEFT", "TOP")
                self["Container_Element_" .. v.Name]:Size(Width, 20 + (((v.Lines or 1) - 1) * 12))
                self["Container_Element_" .. v.Name]:SetChecked(v.Value)
                self["Container_Element_" .. v.Name]:SetDisable(v.Disable)
                self["Container_Element_" .. v.Name].OnChecked = v.Func
            elseif v.Type == "SB" then
                self["Container_Element_" .. v.Name] = self.self:CreateSelectbutton(self.Content)
                if v.Text then 
                    self["Container_Element_" .. v.Name]:Size(Width, 40)
                    self["Container_Element_" .. v.Name]:SetText(v.Text, 11, "Normal", nil, nil, "LEFT", "TOP")
                else 
                    self["Container_Element_" .. v.Name]:Size(Width, 20)
                end
                if type(v.Options) == "function" then
                    self["Container_Element_" .. v.Name].SelectListFunc = v.Options
                else
                    self["Container_Element_" .. v.Name].SelectList = v.Options
                end
                self["Container_Element_" .. v.Name]:SetSelected(v.Value)
                self["Container_Element_" .. v.Name]:SetDisable(v.Disable)
                self["Container_Element_" .. v.Name].OnSelect = v.Func
            elseif v.Type == "NB" then
                self["Container_Element_" .. v.Name] = self.self:CreateSlider(self.Content)
                self["Container_Element_" .. v.Name]:Size(Width, 60)
                self["Container_Element_" .. v.Name]:SetText(v.Text, 11, "Normal", nil, nil, "LEFT", "TOP")
                self["Container_Element_" .. v.Name]:SetDisable(v.Disable)
                self["Container_Element_" .. v.Name]:SetMinStepMax(v.Min, v.Step, v.Max)
                self["Container_Element_" .. v.Name]:SetValue(v.Value)
                self["Container_Element_" .. v.Name].IgnoreSlideUpdates = v.IgnoreSlideUpdates
                self["Container_Element_" .. v.Name]:SetPoint("CENTER")
                self["Container_Element_" .. v.Name].OnValueChanged = v.Func
            elseif v.Type == "CP" then 
                self["Container_Element_" .. v.Name] = self.self:CreateColorSelect(self.Content)
                self["Container_Element_" .. v.Name]:Size(Width, 20)
                self["Container_Element_" .. v.Name]:SetText(v.Text, 11, "Normal", nil, nil, "LEFT", "TOP")
                self["Container_Element_" .. v.Name]:SetDisable(v.Disable)
                self["Container_Element_" .. v.Name]:SetValue(v.r,v.g,v.b,v.a)
                self["Container_Element_" .. v.Name].OnValueChanged = v.Func    
            elseif v.Type == "TB" then
                local BaseHeight = 20
                if v.Simple then 
                    self["Container_Element_" .. v.Name] = self.self:CreateBasicTextBox(self.Content, v.Lines, "LEFT")
                else
                    self["Container_Element_" .. v.Name] = self.self:CreateTextBox(self.Content, v.Lines)
                    self["Container_Element_" .. v.Name]:SetText(v.Text, 11, "Normal", nil, nil, "LEFT", "TOP")
                    BaseHeight = 60
                end
                self["Container_Element_" .. v.Name]:Size(Width, BaseHeight + (type(v.Lines) == "number" and (v.Lines - 1) * 11 or 0))
                self["Container_Element_" .. v.Name]:SetValue(v.Value)
                self["Container_Element_" .. v.Name]:SetDisable(v.Disable)
                self["Container_Element_" .. v.Name].OnValueChanged = v.Func
            elseif v.Type == "SL" then
                self["Container_Element_" .. v.Name] = self.self:CreateSelectList(self.Content)
                self["Container_Element_" .. v.Name]:Size(Width, v.Height)
                self["Container_Element_" .. v.Name]:SetText(v.Text, 11, "Normal", nil, nil, "LEFT", "TOP")
                self["Container_Element_" .. v.Name]:AddItems(v.List)
                self["Container_Element_" .. v.Name]:SetDisable(v.Disable)
                self["Container_Element_" .. v.Name]:SelectItem(v.Selected)
                self["Container_Element_" .. v.Name].OnSelect = v.Func
            elseif v.Type == "Line" then 
                self["Container_Element_" .. v.Name] = self.self:FetchTexture(self.Content)
                self["Container_Element_" .. v.Name]:SetWidth(Width)
                self["Container_Element_" .. v.Name]:SetHeight(v.Thickness or self.self.BorderSize)
                self.self:SetThemedColor(self["Container_Element_" .. v.Name],v.Color or "Border")       
            end

            if self["Container_Element_" .. v.Name] then 
                if PrevFrame then
                    self["Container_Element_" .. v.Name]:SetPoint("TOPLEFT",self["Container_Element_" .. PrevFrame],"TOPRIGHT",self.self.Padding,0)                
                else 
                    if self.ContainerRowData[CurrentRowIndex - 1] and self.ContainerRowData[CurrentRowIndex - 1].SFR then 
                        self["Container_Element_" .. v.Name]:SetPoint("TOPLEFT",self["Container_Element_" .. self.ContainerRowData[CurrentRowIndex - 1].SFR],"TOPLEFT"
                        ,0,-(self.ContainerRowData[CurrentRowIndex - 1].Height + self.self.Padding + HeightOffset - self.ContainerRowData[CurrentRowIndex - 1].HeightOffset))
                    else 
                        self["Container_Element_" .. v.Name]:SetPoint("TOPLEFT",self.self.Padding,-self.self.Padding)
                    end 
                    self.ContainerRowData[CurrentRowIndex].SFR = v.Name 
                end 
                self["Container_Element_" .. v.Name].Tooltip = v.Tooltip;
                if v.Hide then 
                    self["Container_Element_" .. v.Name]:Hide()
                end 
                if self.ContainerRowData[CurrentRowIndex].Height - HeightOffset < self["Container_Element_" .. v.Name]:GetHeight() then 
                    self.ContainerRowData[CurrentRowIndex].Height = self["Container_Element_" .. v.Name]:GetHeight() + HeightOffset
                end 
                PrevFrame = v.Name
                insert(self.ElementNames,PrevFrame)
                self.NumDisplayItems = self.NumDisplayItems + 1
            end 
        end 
    end
    local TotalHeightContent = self.self.Padding
    for i,v in ipairs(self.ContainerRowData) do 
        TotalHeightContent = TotalHeightContent + v.Height + self.self.Padding
    end 
    self.Content:Size(nil,TotalHeightContent)
    return TotalHeightContent
end

local function ChangeItem(self, Item, Par, Value)
    if self["Container_Element_" .. Item] then
        if Par == "Selected" and self["Container_Element_" .. Item].SetFocus then
            self["Container_Element_" .. Item]:SetFocus(Value)
        elseif Par == "Selected" and self["Container_Element_" .. Item].SelectItem then
            self["Container_Element_" .. Item]:SelectItem(Value)
        elseif Par == "Disable" and self["Container_Element_" .. Item].SetDisable then
            self["Container_Element_" .. Item]:SetDisable(Value)
        elseif Par == "Hide" then
            if Value then
                self["Container_Element_" .. Item]:Hide()
            else
                self["Container_Element_" .. Item]:Show()
            end
        elseif Par == "Value" and self["Container_Element_" .. Item].SetValue then
            self["Container_Element_" .. Item]:SetValue(Value)
        elseif Par == "Value" and self["Container_Element_" .. Item].SetChecked then
            self["Container_Element_" .. Item]:SetChecked(Value)
        elseif Par == "Value" and self["Container_Element_" .. Item].SetSelected then
            self["Container_Element_" .. Item]:SetSelected(Value)
        elseif Par == "List" and self["Container_Element_" .. Item].AddItems then
            self["Container_Element_" .. Item]:AddItems(Value)
        elseif Par == "Text" and self["Container_Element_" .. Item].SetText then
            self["Container_Element_" .. Item]:SetText(Value)
        end
    end
end

function _Avyr_Global_ADDON_Elements_Lib.WidgetsMixin:CreateContainer(Parent)
    local Container = self:CreateAdancedScrollFrame(Parent)

    Container.AddRow = AddRow
    Container.ClearItems = ReleaseElements
    Container.ChangeItem = ChangeItem

    Container:SizeChanged(function(self)
        self.Content:Size(self:GetWidth() - 2 * self.self.BorderSize, nil)
    end)

    Container.OnContentWidthChanged = function(self)
        for Rowi,Rowv in ipairs(self.ContainerRowData) do 
            for Elementi,Elementv in ipairs(Rowv.ElementWidthModifier) do 
                local Width = 0
                if Elementv.FixedWidth then 
                    Width = Elementv.FixedWidth
                else
                    Width = Elementv.Ratio * (self.Content:GetWidth() - (getn(Rowv.ElementWidthModifier) + 1) * self.self.Padding - Elementv.RatioOffset)
                end
                if self["Container_Element_" .. Elementv.Name] then 
                    if self["Container_Element_" .. Elementv.Name].Size then 
                        self["Container_Element_" .. Elementv.Name]:Size(Width)
                    else
                        self["Container_Element_" .. Elementv.Name]:SetWidth(Width)
                    end 
                end
            end
        end 
    end

    Container.ScrollDistance = 0
    Container.NumDisplayItems = 0
    Container.ContainerRowData = {}
    Container.ElementNames = {}

    Container.SetScroll = function(self, Value, Increment)
        Value = Value or self.ScrollDistance
        if Increment then
            self.ScrollDistance = self.ScrollDistance + Value
        else
            self.ScrollDistance = Value
        end
        if self.ScrollDistance < 0 then
            self.ScrollDistance = 0
        elseif self.ScrollDistance > self.Content:GetHeight() - self.ScrollFrame:GetHeight() then
            if self.Content:GetHeight() - self.ScrollFrame:GetHeight() > 0 then
                self.ScrollDistance = self.Content:GetHeight() - self.ScrollFrame:GetHeight()
            else
                self.ScrollDistance = 0
            end
        end
        self.ScrollFrame:SetVerticalScroll(self.ScrollDistance)
    end
    return Container
end
